@@ -524,7 +524,7 @@ end
 
 ## world view of the cache
 if !isdefined(CC, :WorldView)
-    using Core.Compiler: InternalCodeCache as WorldView
+    using Core.Compiler: InternalCodeCache
 else
     using Core.Compiler: WorldView
 end
@@ -649,7 +649,17 @@ function ci_cache_populate(interp, cache, mi, min_world, max_world)
 end
 
 function ci_cache_lookup(cache, mi, min_world, max_world)
-    wvc = WorldView(cache, min_world, max_world)
+    @static if !isdefined(CC, :WorldView)
+        # Use the overlay cache if possible
+        globalcache = cache.globalcache
+        cacheworlds = globalcache.worlds
+        if cacheworlds.min_world == min_world && cacheworlds.min_world == max_world
+            return CC.get(cache, mi, nothing)
+        end
+        wvc = InternalCodeCache(globalcache.owner, min_world, max_world)
+    else
+        wvc = WorldView(cache, min_world, max_world)
+    end
     ci = CC.get(wvc, mi, nothing)
     if VERSION < v"1.12.0-DEV.1434" && ci !== nothing && ci.inferred === nothing
         # if for some reason we did end up with a codeinfo without inferred source, e.g.,
